﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using Microsoft.Kinect;

namespace BeginningKinect.Chapter4.KinectTheDots
{
    public class KinectViewModel
    {
        private KinectSensor mKinectDevice;
        private Skeleton[] mFrameSkeletons;
        private bool mIsHandVisible;
        private int mHandDirection;

        public KinectViewModel()
        {
            KinectSensor.KinectSensors.StatusChanged += new EventHandler<StatusChangedEventArgs>(KinectSensors_StatusChanged);
            KinectDevice = KinectSensor.KinectSensors.FirstOrDefault(x => x.Status == KinectStatus.Connected);
        }

        void KinectSensors_StatusChanged(object sender, StatusChangedEventArgs e)
        {
            switch (e.Status)
            {
                case KinectStatus.Initializing:
                case KinectStatus.Connected:
                case KinectStatus.NotPowered:
                case KinectStatus.NotReady:
                case KinectStatus.DeviceNotGenuine:
                    this.KinectDevice = e.Sensor;
                    break;
                case KinectStatus.Disconnected:
                    //TODO: Give the user feedback to plug-in a Kinect device.                    
                    this.KinectDevice = null;
                    break;
                default:
                    //TODO: Show an error state
                    break;
            }
        }

        private void KinectDevice_SkeletonFrameReady(object sender, SkeletonFrameReadyEventArgs e)
        {
            using (SkeletonFrame frame = e.OpenSkeletonFrame())
            {
                if (frame != null)
                {
                    frame.CopySkeletonDataTo(mFrameSkeletons);
                    Skeleton skeleton = GetPrimarySkeleton(mFrameSkeletons);

                    if (skeleton == null)
                    {
                        IsHandVisible = false;
                    }
                    else
                    {                        
                        Joint primaryHand = GetPrimaryHand(skeleton);
                        TrackHand(primaryHand);
                    }
                }
            }
        }

        public bool IsHandVisible 
        {
            get { return mIsHandVisible; }
            set { mIsHandVisible = value; } 
        }

        public KinectSensor KinectDevice
        {
            get { return mKinectDevice; }
            set
            {
                if (mKinectDevice != value)
                {
                    //Uninitialize
                    if (mKinectDevice != null)
                    {
                        mKinectDevice.Stop();
                        mKinectDevice.SkeletonFrameReady -= KinectDevice_SkeletonFrameReady;
                        mKinectDevice.SkeletonStream.Disable();
                        mFrameSkeletons = null;
                    }

                    mKinectDevice = value;

                    //Initialize
                    if (mKinectDevice != null)
                    {
                        if (mKinectDevice.Status == KinectStatus.Connected)
                        {
                            mKinectDevice.SkeletonStream.Enable();
                            mFrameSkeletons = new Skeleton[mKinectDevice.SkeletonStream.FrameSkeletonArrayLength];
                            mKinectDevice.Start();                            
                            KinectDevice.SkeletonFrameReady += KinectDevice_SkeletonFrameReady;
                        }
                    }
                }
            }
        }

        
        public int HandDirection
        {
            get { return mHandDirection; }
            set { mHandDirection = value; }
        }

        private static Skeleton GetPrimarySkeleton(Skeleton[] skeletons)
        {
            Skeleton skeleton = null;

            if (skeletons != null)
            {
                //Find the closest skeleton       
                for (int i = 0; i < skeletons.Length; i++)
                {
                    if (skeletons[i].TrackingState == SkeletonTrackingState.Tracked)
                    {
                        if (skeleton == null)
                        {
                            skeleton = skeletons[i];
                        }
                        else
                        {
                            if (skeleton.Position.Z > skeletons[i].Position.Z)
                            {
                                skeleton = skeletons[i];
                            }
                        }
                    }
                }
            }

            return skeleton;
        }

        private static Joint GetPrimaryHand(Skeleton skeleton)
        {
            Joint primaryHand = new Joint();
            
            if (skeleton != null)
            {
                primaryHand = skeleton.Joints[JointType.HandLeft];
                Joint rightHand = skeleton.Joints[JointType.HandRight];                
                if (rightHand.TrackingState != JointTrackingState.NotTracked)
                {
                    if (primaryHand.TrackingState == JointTrackingState.NotTracked)
                    {
                        primaryHand = rightHand;                        
                    }
                    else
                    {
                        if (primaryHand.Position.Z > rightHand.Position.Z)
                        {
                            primaryHand = rightHand;
                        }
                    }
                }                
            }

            return primaryHand;
        }


        private void TrackHand(Joint hand)
        {
            if (hand.TrackingState == JointTrackingState.NotTracked)
            {
                IsHandVisible = false;
            }
            else
            {
                IsHandVisible = true;


                DepthImagePoint point = mKinectDevice.MapSkeletonPointToDepth(hand.Position, mKinectDevice.DepthStream.Format);
                //point.X = (int)((point.X * LayoutRoot.ActualWidth / _KinectDevice.DepthStream.FrameWidth) - (HandCursorElement.ActualWidth / 2.0));
                //point.Y = (int)((point.Y * LayoutRoot.ActualHeight / _KinectDevice.DepthStream.FrameHeight) - (HandCursorElement.ActualHeight / 2.0));

                //Canvas.SetLeft(HandCursorElement, point.X);
                //Canvas.SetTop(HandCursorElement, point.Y);

                if (hand.JointType == JointType.HandRight)
                {
                    HandDirection = 1;
                }
                else
                {
                    HandDirection = -1;
                }
            }
        }

    }
}
